@@ -13,8 +13,6 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-__all__ = ['AMQPDriverBase']
-
 import logging
 import threading
 import time
@@ -28,10 +26,9 @@ import oslo_messaging
 from oslo_messaging._drivers import amqp as rpc_amqp
 from oslo_messaging._drivers import base
 from oslo_messaging._drivers import common as rpc_common
-from oslo_messaging._i18n import _
-from oslo_messaging._i18n import _LE
-from oslo_messaging._i18n import _LI
-from oslo_messaging._i18n import _LW
+from oslo_messaging import MessageDeliveryFailure
+
+__all__ = ['AMQPDriverBase']
 
 LOG = logging.getLogger(__name__)
 
@@ -42,32 +39,67 @@ ACK_REQUEUE_EVERY_SECONDS_MIN = 0.001
 ACK_REQUEUE_EVERY_SECONDS_MAX = 1.0
 
 
-def do_pending_tasks(tasks):
-    while True:
-        try:
-            task = tasks.get(block=False)
-        except moves.queue.Empty:
-            break
-        else:
+class MessageOperationsHandler(object):
+    """Queue used by message operations to ensure that all tasks are
+    serialized and run in the same thread, since underlying drivers like kombu
+    are not thread safe.
+    """
+    def __init__(self, name):
+        self.name = "%s (%s)" % (name, hex(id(self)))
+        self._tasks = moves.queue.Queue()
+
+        self._shutdown = threading.Event()
+        self._shutdown_thread = threading.Thread(
+            target=self._process_in_background)
+        self._shutdown_thread.daemon = True
+
+    def stop(self):
+        self._shutdown.set()
+
+    def process_in_background(self):
+        """Run all pending tasks queued by do() in an thread during the
+        shutdown process.
+        """
+        self._shutdown_thread.start()
+
+    def _process_in_background(self):
+        while not self._shutdown.is_set():
+            self.process()
+            time.sleep(ACK_REQUEUE_EVERY_SECONDS_MIN)
+
+    def process(self):
+        "Run all pending tasks queued by do()."
+
+        while True:
+            try:
+                task = self._tasks.get(block=False)
+            except moves.queue.Empty:
+                break
             task()
 
+    def do(self, task):
+        "Put the task in the queue."
+        self._tasks.put(task)
+
 
 class AMQPIncomingMessage(base.RpcIncomingMessage):
 
     def __init__(self, listener, ctxt, message, unique_id, msg_id, reply_q,
-                 obsolete_reply_queues, pending_message_actions):
+                 client_timeout, obsolete_reply_queues,
+                 message_operations_handler):
         super(AMQPIncomingMessage, self).__init__(ctxt, message)
         self.listener = listener
 
         self.unique_id = unique_id
         self.msg_id = msg_id
         self.reply_q = reply_q
+        self.client_timeout = client_timeout
         self._obsolete_reply_queues = obsolete_reply_queues
-        self._pending_tasks = pending_message_actions
+        self._message_operations_handler = message_operations_handler
         self.stopwatch = timeutils.StopWatch()
         self.stopwatch.start()
 
-    def _send_reply(self, conn, reply=None, failure=None):
+    def _send_reply(self, conn, reply=None, failure=None, ending=True):
         if not self._obsolete_reply_queues.reply_q_valid(self.reply_q,
                                                          self.msg_id):
             return
@@ -76,7 +108,7 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
             failure = rpc_common.serialize_remote_exception(failure)
         # NOTE(sileht): ending can be removed in N*, see Listener.wait()
         # for more detail.
-        msg = {'result': reply, 'failure': failure, 'ending': True,
+        msg = {'result': reply, 'failure': failure, 'ending': ending,
                '_msg_id': self.msg_id}
         rpc_amqp._add_unique_id(msg)
         unique_id = msg[rpc_amqp.UNIQUE_ID]
@@ -124,16 +156,50 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
                     time.sleep(0.25)
                 else:
                     self._obsolete_reply_queues.add(self.reply_q, self.msg_id)
-                    LOG.info(_LI("The reply %(msg_id)s cannot be sent  "
-                                 "%(reply_q)s reply queue don't exist after "
-                                 "%(duration)s sec abandoning..."), {
-                                     'msg_id': self.msg_id,
-                                     'reply_q': self.reply_q,
-                                     'duration': duration})
+                    infos = {
+                        'msg_id': self.msg_id,
+                        'reply_q': self.reply_q,
+                        'duration': duration
+                    }
+                    LOG.info("The reply %(msg_id)s cannot be sent  "
+                             "%(reply_q)s reply queue don't exist after "
+                             "%(duration)s sec abandoning...", infos)
                     return
 
+    def heartbeat(self):
+        # generate a keep alive for RPC call monitoring
+        with self.listener.driver._get_connection(
+                rpc_common.PURPOSE_SEND) as conn:
+            try:
+                self._send_reply(conn, None, None, ending=False)
+            except rpc_amqp.AMQPDestinationNotFound:
+                # internal exception that indicates queue/exchange gone -
+                # broker unreachable.
+                raise MessageDeliveryFailure("Heartbeat send failed")
+
+    # NOTE(sileht): Those have already be ack in RpcListener IO thread
+    # We keep them as noop until all drivers do the same
     def acknowledge(self):
-        self._pending_tasks.put(self.message.acknowledge)
+        pass
+
+    def requeue(self):
+        pass
+
+
+class NotificationAMQPIncomingMessage(AMQPIncomingMessage):
+    def acknowledge(self):
+        def _do_ack():
+            try:
+                self.message.acknowledge()
+            except Exception as exc:
+                # NOTE(kgiusti): this failure is likely due to a loss of the
+                # connection to the broker.  Not much we can do in this case,
+                # especially considering the Notification has already been
+                # dispatched. This *could* result in message duplication
+                # (unacked msg is returned to the queue by the broker), but the
+                # driver tries to catch that using the msg_id_cache.
+                LOG.warning("Failed to acknowledge received message: %s", exc)
+        self._message_operations_handler.do(_do_ack)
         self.listener.msg_id_cache.add(self.unique_id)
 
     def requeue(self):
@@ -143,7 +209,12 @@ class AMQPIncomingMessage(base.RpcIncomingMessage):
         # msg_id_cache, the message will be reconsumed, the only difference is
         # the message stay at the beginning of the queue instead of moving to
         # the end.
-        self._pending_tasks.put(self.message.requeue)
+        def _do_requeue():
+            try:
+                self.message.requeue()
+            except Exception as exc:
+                LOG.warning("Failed to requeue received message: %s", exc)
+        self._message_operations_handler.do(_do_requeue)
 
 
 class ObsoleteReplyQueuesCache(object):
@@ -186,9 +257,8 @@ class ObsoleteReplyQueuesCache(object):
         self._no_reply_log(reply_q, msg_id)
 
     def _no_reply_log(self, reply_q, msg_id):
-        LOG.warning(_LW("%(reply_queue)s doesn't exists, drop reply to "
-                        "%(msg_id)s"), {'reply_queue': reply_q,
-                                        'msg_id': msg_id})
+        LOG.warning("%(reply_queue)s doesn't exists, drop reply to "
+                    "%(msg_id)s", {'reply_queue': reply_q, "msg_id": msg_id})
 
 
 class AMQPListener(base.PollStyleListener):
@@ -199,9 +269,11 @@ class AMQPListener(base.PollStyleListener):
         self.conn = conn
         self.msg_id_cache = rpc_amqp._MsgIdCache()
         self.incoming = []
-        self._stopped = threading.Event()
+        self._shutdown = threading.Event()
+        self._shutoff = threading.Event()
         self._obsolete_reply_queues = ObsoleteReplyQueuesCache()
-        self._pending_tasks = moves.queue.Queue()
+        self._message_operations_handler = MessageOperationsHandler(
+            "AMQPListener")
         self._current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
 
     def __call__(self, message):
@@ -214,22 +286,23 @@ class AMQPListener(base.PollStyleListener):
         else:
             LOG.debug("received message with unique_id: %s", unique_id)
 
-        self.incoming.append(AMQPIncomingMessage(
+        self.incoming.append(self.message_cls(
             self,
             ctxt.to_dict(),
             message,
             unique_id,
             ctxt.msg_id,
             ctxt.reply_q,
+            ctxt.client_timeout,
             self._obsolete_reply_queues,
-            self._pending_tasks))
+            self._message_operations_handler))
 
     @base.batch_poll_helper
     def poll(self, timeout=None):
         stopwatch = timeutils.StopWatch(duration=timeout).start()
 
-        while not self._stopped.is_set():
-            do_pending_tasks(self._pending_tasks)
+        while not self._shutdown.is_set():
+            self._message_operations_handler.process()
 
             if self.incoming:
                 return self.incoming.pop(0)
@@ -248,16 +321,69 @@ class AMQPListener(base.PollStyleListener):
             else:
                 self._current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
 
+        # NOTE(sileht): listener is stopped, just processes remaining messages
+        # and operations
+        self._message_operations_handler.process()
+        if self.incoming:
+            return self.incoming.pop(0)
+
+        self._shutoff.set()
+
     def stop(self):
-        self._stopped.set()
+        self._shutdown.set()
         self.conn.stop_consuming()
-        do_pending_tasks(self._pending_tasks)
+        self._shutoff.wait()
+
+        # NOTE(sileht): Here, the listener is stopped, but some incoming
+        # messages may still live on server side, because callback is still
+        # running and message is not yet ack/requeue. It's safe to do the ack
+        # into another thread, side the polling thread is now terminated.
+        self._message_operations_handler.process_in_background()
 
     def cleanup(self):
+        # NOTE(sileht): server executor is now stopped, we are sure that no
+        # more incoming messages in live, we can acknowledge
+        # remaining messages and stop the thread
+        self._message_operations_handler.stop()
         # Closes listener connection
         self.conn.close()
 
 
+class RpcAMQPListener(AMQPListener):
+    message_cls = AMQPIncomingMessage
+
+    def __call__(self, message):
+        # NOTE(kgiusti): In the original RPC implementation the RPC server
+        # would acknowledge the request THEN process it.  The goal of this was
+        # to prevent duplication if the ack failed.  Should the ack fail the
+        # request would be discarded since the broker would not remove the
+        # request from the queue since no ack was received.  That would lead to
+        # the request being redelivered at some point. However this approach
+        # meant that the ack was issued from the dispatch thread, not the
+        # consumer thread, which is bad since kombu is not thread safe.  So a
+        # change was made to schedule the ack to be sent on the consumer thread
+        # - breaking the ability to catch ack errors before dispatching the
+        # request.  To fix this we do the actual ack here in the consumer
+        # callback and avoid the upcall if the ack fails.  See
+        # https://bugs.launchpad.net/oslo.messaging/+bug/1695746
+        # for all the gory details...
+        try:
+            message.acknowledge()
+        except Exception as exc:
+            LOG.warning("Discarding RPC request due to failed acknowledge: %s",
+                        exc)
+        else:
+            # NOTE(kgiusti): be aware that even if the acknowledge call
+            # succeeds there is no guarantee the broker actually gets the ACK
+            # since acknowledge() simply writes the ACK to the socket (there is
+            # no ACK confirmation coming back from the broker)
+            super(RpcAMQPListener, self).__call__(message)
+
+
+class NotificationAMQPListener(AMQPListener):
+    message_cls = NotificationAMQPIncomingMessage
+
+
 class ReplyWaiters(object):
 
     WAKE_UP = object()
@@ -277,7 +403,7 @@ class ReplyWaiters(object):
     def put(self, msg_id, message_data):
         queue = self._queues.get(msg_id)
         if not queue:
-            LOG.info(_LI('No calling threads waiting for msg_id : %s'), msg_id)
+            LOG.info('No calling threads waiting for msg_id : %s', msg_id)
             LOG.debug(' queues: %(queues)s, message: %(message)s',
                       {'queues': len(self._queues), 'message': message_data})
         else:
@@ -287,10 +413,10 @@ class ReplyWaiters(object):
         self._queues[msg_id] = moves.queue.Queue()
         queues_length = len(self._queues)
         if queues_length > self._wrn_threshold:
-            LOG.warning(_LW('Number of call queues is %(queues_length)s, '
-                            'greater than warning threshold: %(old_threshold)s'
-                            '. There could be a leak. Increasing threshold to:'
-                            ' %(threshold)s'),
+            LOG.warning('Number of call queues is %(queues_length)s, '
+                        'greater than warning threshold: %(old_threshold)s. '
+                        'There could be a leak. Increasing threshold to: '
+                        '%(threshold)s',
                         {'queues_length': queues_length,
                          'old_threshold': self._wrn_threshold,
                          'threshold': self._wrn_threshold * 2})
@@ -306,7 +432,6 @@ class ReplyWaiter(object):
         self.allowed_remote_exmods = allowed_remote_exmods
         self.msg_id_cache = rpc_amqp._MsgIdCache()
         self.waiters = ReplyWaiters()
-        self._pending_tasks = moves.queue.Queue()
 
         self.conn.declare_direct_consumer(reply_q, self)
 
@@ -321,12 +446,10 @@ class ReplyWaiter(object):
             self.conn.stop_consuming()
             self._thread.join()
             self._thread = None
-            do_pending_tasks(self._pending_tasks)
 
     def poll(self):
         current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
         while not self._thread_exit_event.is_set():
-            do_pending_tasks(self._pending_tasks)
             try:
                 # ack every ACK_REQUEUE_EVERY_SECONDS_MAX seconds
                 self.conn.consume(timeout=current_timeout)
@@ -334,13 +457,16 @@ class ReplyWaiter(object):
                 current_timeout = max(current_timeout * 2,
                                       ACK_REQUEUE_EVERY_SECONDS_MAX)
             except Exception:
-                LOG.exception(_LE("Failed to process incoming message, "
-                              "retrying..."))
+                LOG.exception("Failed to process incoming message, retrying..")
             else:
                 current_timeout = ACK_REQUEUE_EVERY_SECONDS_MIN
 
     def __call__(self, message):
-        self._pending_tasks.put(message.acknowledge)
+        # NOTE(sileht): __call__ is running within the polling thread,
+        # (conn.consume -> conn.conn.drain_events() -> __call__ callback)
+        # it's threadsafe to acknowledge the message here, no need to wait
+        # the next polling
+        message.acknowledge()
         incoming_msg_id = message.pop('_msg_id', None)
         if message.get('ending'):
             LOG.debug("received reply msg_id: %s", incoming_msg_id)
@@ -355,7 +481,7 @@ class ReplyWaiter(object):
     @staticmethod
     def _raise_timeout_exception(msg_id):
         raise oslo_messaging.MessagingTimeout(
-            _('Timed out waiting for a reply to message ID %s.') % msg_id)
+            'Timed out waiting for a reply to message ID %s.', msg_id)
 
     def _process_reply(self, data):
         self.msg_id_cache.check_duplicate_message(data)
@@ -369,7 +495,7 @@ class ReplyWaiter(object):
         ending = data.get('ending', False)
         return result, ending
 
-    def wait(self, msg_id, timeout):
+    def wait(self, msg_id, timeout, call_monitor_timeout):
         # NOTE(sileht): for each msg_id we receive two amqp message
         # first one with the payload, a second one to ensure the other
         # have finish to send the payload
@@ -378,10 +504,21 @@ class ReplyWaiter(object):
         # support both cases for now.
         timer = rpc_common.DecayingTimer(duration=timeout)
         timer.start()
+        if call_monitor_timeout:
+            call_monitor_timer = rpc_common.DecayingTimer(
+                duration=call_monitor_timeout)
+            call_monitor_timer.start()
+        else:
+            call_monitor_timer = None
         final_reply = None
         ending = False
         while not ending:
             timeout = timer.check_return(self._raise_timeout_exception, msg_id)
+            if call_monitor_timer and timeout > 0:
+                cm_timeout = call_monitor_timer.check_return(
+                    self._raise_timeout_exception, msg_id)
+                if cm_timeout < timeout:
+                    timeout = cm_timeout
             try:
                 message = self.waiters.get(msg_id, timeout=timeout)
             except moves.queue.Empty:
@@ -393,6 +530,10 @@ class ReplyWaiter(object):
                 # empty `result` field or a second _send_reply() with
                 # ending=True and no `result` field.
                 final_reply = reply
+            elif ending is False:
+                LOG.debug('Call monitor heartbeat received; '
+                          'renewing timeout timer')
+                call_monitor_timer.restart()
         return final_reply
 
 
@@ -438,8 +579,8 @@ class AMQPDriverBase(base.BaseDriver):
         return self._reply_q
 
     def _send(self, target, ctxt, message,
-              wait_for_reply=None, timeout=None,
-              envelope=True, notify=False, retry=None):
+              wait_for_reply=None, timeout=None, call_monitor_timeout=None,
+              envelope=True, notify=False, retry=None, transport_options=None):
 
         msg = message
 
@@ -447,6 +588,7 @@ class AMQPDriverBase(base.BaseDriver):
             msg_id = uuid.uuid4().hex
             msg.update({'_msg_id': msg_id})
             msg.update({'_reply_q': self._get_reply_q()})
+            msg.update({'_timeout': call_monitor_timeout})
 
         rpc_amqp._add_unique_id(msg)
         unique_id = msg[rpc_amqp.UNIQUE_ID]
@@ -466,11 +608,9 @@ class AMQPDriverBase(base.BaseDriver):
             with self._get_connection(rpc_common.PURPOSE_SEND) as conn:
                 if notify:
                     exchange = self._get_exchange(target)
-                    log_msg += "NOTIFY exchange '%(exchange)s'" \
-                               " topic '%(topic)s'" % {
-                                   'exchange': exchange,
-                                   'topic': target.topic}
-                    LOG.debug(log_msg)
+                    LOG.debug(log_msg + "NOTIFY exchange '%(exchange)s'"
+                              " topic '%(topic)s'", {'exchange': exchange,
+                                                     'topic': target.topic})
                     conn.notify_send(exchange, target.topic, msg, retry=retry)
                 elif target.fanout:
                     log_msg += "FANOUT topic '%(topic)s'" % {
@@ -482,16 +622,16 @@ class AMQPDriverBase(base.BaseDriver):
                     exchange = self._get_exchange(target)
                     if target.server:
                         topic = '%s.%s' % (target.topic, target.server)
-                    log_msg += "exchange '%(exchange)s'" \
-                               " topic '%(topic)s'" % {
-                                   'exchange': exchange,
-                                   'topic': topic}
-                    LOG.debug(log_msg)
+                    LOG.debug(log_msg + "exchange '%(exchange)s'"
+                              " topic '%(topic)s'", {'exchange': exchange,
+                                                     'topic': topic})
                     conn.topic_send(exchange_name=exchange, topic=topic,
-                                    msg=msg, timeout=timeout, retry=retry)
+                                    msg=msg, timeout=timeout, retry=retry,
+                                    transport_options=transport_options)
 
             if wait_for_reply:
-                result = self._waiter.wait(msg_id, timeout)
+                result = self._waiter.wait(msg_id, timeout,
+                                           call_monitor_timeout)
                 if isinstance(result, Exception):
                     raise result
                 return result
@@ -500,9 +640,10 @@ class AMQPDriverBase(base.BaseDriver):
                 self._waiter.unlisten(msg_id)
 
     def send(self, target, ctxt, message, wait_for_reply=None, timeout=None,
-             retry=None):
+             call_monitor_timeout=None, retry=None, transport_options=None):
         return self._send(target, ctxt, message, wait_for_reply, timeout,
-                          retry=retry)
+                          call_monitor_timeout, retry=retry,
+                          transport_options=transport_options)
 
     def send_notification(self, target, ctxt, message, version, retry=None):
         return self._send(target, ctxt, message,
@@ -511,7 +652,7 @@ class AMQPDriverBase(base.BaseDriver):
     def listen(self, target, batch_size, batch_timeout):
         conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
 
-        listener = AMQPListener(self, conn)
+        listener = RpcAMQPListener(self, conn)
 
         conn.declare_topic_consumer(exchange_name=self._get_exchange(target),
                                     topic=target.topic,
@@ -528,14 +669,8 @@ class AMQPDriverBase(base.BaseDriver):
     def listen_for_notifications(self, targets_and_priorities, pool,
                                  batch_size, batch_timeout):
         conn = self._get_connection(rpc_common.PURPOSE_LISTEN)
-        # NOTE(sileht): The application set batch_size, so we don't need to
-        # prefetch more messages, especially for notifications. Notifications
-        # queues can be really big when the consumer have disapear during a
-        # long period, and when it come back, kombu/pyamqp will fetch all
-        # messages it can. So we override the default qos prefetch value
-        conn.connection.rabbit_qos_prefetch_count = batch_size
 
-        listener = AMQPListener(self, conn)
+        listener = NotificationAMQPListener(self, conn)
         for target, priority in targets_and_priorities:
             conn.declare_topic_consumer(
                 exchange_name=self._get_exchange(target),
