@@ -24,13 +24,13 @@ import oslo_messaging
 from oslo_messaging._drivers import base
 
 
-class FakeIncomingMessage(base.IncomingMessage):
-    def __init__(self, listener, ctxt, message, reply_q, requeue):
-        super(FakeIncomingMessage, self).__init__(listener, ctxt, message)
+class FakeIncomingMessage(base.RpcIncomingMessage):
+    def __init__(self, ctxt, message, reply_q, requeue):
+        super(FakeIncomingMessage, self).__init__(ctxt, message)
         self.requeue_callback = requeue
         self._reply_q = reply_q
 
-    def reply(self, reply=None, failure=None, log_failure=True):
+    def reply(self, reply=None, failure=None):
         if self._reply_q:
             failure = failure[1] if failure else None
             self._reply_q.put((reply, failure))
@@ -38,11 +38,14 @@ class FakeIncomingMessage(base.IncomingMessage):
     def requeue(self):
         self.requeue_callback()
 
+    def heartbeat(self):
+        """Heartbeat is not supported."""
 
-class FakeListener(base.Listener):
 
-    def __init__(self, driver, exchange_manager, targets, pool=None):
-        super(FakeListener, self).__init__(driver)
+class FakeListener(base.PollStyleListener):
+
+    def __init__(self, exchange_manager, targets, pool=None):
+        super(FakeListener, self).__init__()
         self._exchange_manager = exchange_manager
         self._targets = targets
         self._pool = pool
@@ -54,6 +57,7 @@ class FakeListener(base.Listener):
             exchange = self._exchange_manager.get_exchange(target.exchange)
             exchange.ensure_queue(target, pool)
 
+    @base.batch_poll_helper
     def poll(self, timeout=None):
         if timeout is not None:
             deadline = time.time() + timeout
@@ -65,8 +69,8 @@ class FakeListener(base.Listener):
                 (ctxt, message, reply_q, requeue) = exchange.poll(target,
                                                                   self._pool)
                 if message is not None:
-                    message = FakeIncomingMessage(self, ctxt, message,
-                                                  reply_q, requeue)
+                    message = FakeIncomingMessage(ctxt, message, reply_q,
+                                                  requeue)
                     return message
             if deadline is not None:
                 pause = deadline - time.time()
@@ -141,10 +145,11 @@ class FakeExchange(object):
 
 
 class FakeExchangeManager(object):
+    _exchanges_lock = threading.Lock()
+    _exchanges = {}
+
     def __init__(self, default_exchange):
         self._default_exchange = default_exchange
-        self._exchanges_lock = threading.Lock()
-        self._exchanges = {}
 
     def get_exchange(self, name):
         if name is None:
@@ -183,7 +188,8 @@ class FakeDriver(base.BaseDriver):
         """
         json.dumps(message)
 
-    def _send(self, target, ctxt, message, wait_for_reply=None, timeout=None):
+    def _send(self, target, ctxt, message, wait_for_reply=None, timeout=None,
+              transport_options=None):
         self._check_serialize(message)
 
         exchange = self._exchange_manager.get_exchange(target.exchange)
@@ -211,19 +217,20 @@ class FakeDriver(base.BaseDriver):
         return None
 
     def send(self, target, ctxt, message, wait_for_reply=None, timeout=None,
-             retry=None):
+             call_monitor_timeout=None, retry=None, transport_options=None):
         # NOTE(sileht): retry doesn't need to be implemented, the fake
         # transport always works
-        return self._send(target, ctxt, message, wait_for_reply, timeout)
+        return self._send(target, ctxt, message, wait_for_reply, timeout,
+                          transport_options)
 
     def send_notification(self, target, ctxt, message, version, retry=None):
         # NOTE(sileht): retry doesn't need to be implemented, the fake
         # transport always works
         self._send(target, ctxt, message)
 
-    def listen(self, target):
+    def listen(self, target, batch_size, batch_timeout):
         exchange = target.exchange or self._default_exchange
-        listener = FakeListener(self, self._exchange_manager,
+        listener = FakeListener(self._exchange_manager,
                                 [oslo_messaging.Target(
                                     topic=target.topic,
                                     server=target.server,
@@ -231,17 +238,20 @@ class FakeDriver(base.BaseDriver):
                                  oslo_messaging.Target(
                                      topic=target.topic,
                                      exchange=exchange)])
-        return listener
+        return base.PollStyleListenerAdapter(listener, batch_size,
+                                             batch_timeout)
 
-    def listen_for_notifications(self, targets_and_priorities, pool):
+    def listen_for_notifications(self, targets_and_priorities, pool,
+                                 batch_size, batch_timeout):
         targets = [
             oslo_messaging.Target(
                 topic='%s.%s' % (target.topic, priority),
                 exchange=target.exchange)
             for target, priority in targets_and_priorities]
-        listener = FakeListener(self, self._exchange_manager, targets, pool)
+        listener = FakeListener(self._exchange_manager, targets, pool)
 
-        return listener
+        return base.PollStyleListenerAdapter(listener, batch_size,
+                                             batch_timeout)
 
     def cleanup(self):
         pass
