@@ -22,39 +22,21 @@ import uuid
 import fixtures
 import kombu
 import kombu.transport.memory
-from oslo_config import cfg
 from oslo_serialization import jsonutils
-from oslotest import mockpatch
 import testscenarios
 
 import oslo_messaging
 from oslo_messaging._drivers import amqpdriver
 from oslo_messaging._drivers import common as driver_common
 from oslo_messaging._drivers import impl_rabbit as rabbit_driver
+from oslo_messaging.exceptions import MessageDeliveryFailure
 from oslo_messaging.tests import utils as test_utils
+from oslo_messaging.transport import DriverLoadFailure
 from six.moves import mock
 
 load_tests = testscenarios.load_tests_apply_scenarios
 
 
-class TestDeprecatedRabbitDriverLoad(test_utils.BaseTestCase):
-
-    def setUp(self):
-        super(TestDeprecatedRabbitDriverLoad, self).setUp(
-            conf=cfg.ConfigOpts())
-        self.messaging_conf.transport_driver = 'rabbit'
-        self.config(fake_rabbit=True, group="oslo_messaging_rabbit")
-
-    def test_driver_load(self):
-        transport = oslo_messaging.get_transport(self.conf)
-        self.addCleanup(transport.cleanup)
-        driver = transport._driver
-        url = driver._get_connection()._url
-
-        self.assertIsInstance(driver, rabbit_driver.RabbitDriver)
-        self.assertEqual('memory:////', url)
-
-
 class TestHeartbeat(test_utils.BaseTestCase):
 
     @mock.patch('oslo_messaging._drivers.impl_rabbit.LOG')
@@ -90,11 +72,11 @@ class TestHeartbeat(test_utils.BaseTestCase):
 
         if not heartbeat_side_effect:
             self.assertEqual(1, fake_ensure_connection.call_count)
-            self.assertEqual(3, fake_logger.debug.call_count)
+            self.assertEqual(2, fake_logger.debug.call_count)
             self.assertEqual(0, fake_logger.info.call_count)
         else:
             self.assertEqual(2, fake_ensure_connection.call_count)
-            self.assertEqual(3, fake_logger.debug.call_count)
+            self.assertEqual(2, fake_logger.debug.call_count)
             self.assertEqual(1, fake_logger.info.call_count)
             self.assertIn(mock.call(info, mock.ANY),
                           fake_logger.info.mock_calls)
@@ -104,10 +86,15 @@ class TestHeartbeat(test_utils.BaseTestCase):
 
     def test_test_heartbeat_sent_connection_fail(self):
         self._do_test_heartbeat_sent(
-            heartbeat_side_effect=kombu.exceptions.ConnectionError,
+            heartbeat_side_effect=kombu.exceptions.OperationalError,
             info='A recoverable connection/channel error occurred, '
             'trying to reconnect: %s')
 
+    def test_run_heartbeat_in_pthread(self):
+        self.config(heartbeat_in_pthread=True,
+                    group="oslo_messaging_rabbit")
+        self._do_test_heartbeat_sent()
+
 
 class TestRabbitQos(test_utils.BaseTestCase):
 
@@ -137,41 +124,51 @@ class TestRabbitQos(test_utils.BaseTestCase):
 class TestRabbitDriverLoad(test_utils.BaseTestCase):
 
     scenarios = [
-        ('rabbit', dict(transport_driver='rabbit',
-                        url='amqp://guest:guest@localhost:5672//')),
-        ('kombu', dict(transport_driver='kombu',
-                       url='amqp://guest:guest@localhost:5672//')),
-        ('rabbit+memory', dict(transport_driver='kombu+memory',
-                               url='memory:///'))
+        ('rabbit', dict(transport_url='rabbit:/guest:guest@localhost:5672//')),
+        ('kombu', dict(transport_url='kombu:/guest:guest@localhost:5672//')),
+        ('rabbit+memory', dict(transport_url='kombu+memory:/'))
     ]
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.reset')
     def test_driver_load(self, fake_ensure, fake_reset):
         self.config(heartbeat_timeout_threshold=60,
                     group='oslo_messaging_rabbit')
-        self.messaging_conf.transport_driver = self.transport_driver
+        self.messaging_conf.transport_url = self.transport_url
         transport = oslo_messaging.get_transport(self.conf)
         self.addCleanup(transport.cleanup)
         driver = transport._driver
-        url = driver._get_connection()._url
 
         self.assertIsInstance(driver, rabbit_driver.RabbitDriver)
-        self.assertEqual(self.url, url)
+
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.reset')
+    def test_driver_load_max_less_than_min(self, fake_ensure, fake_reset):
+        self.config(
+            rpc_conn_pool_size=1, conn_pool_min_size=2,
+            group='oslo_messaging_rabbit')
+        self.messaging_conf.transport_url = self.transport_url
+        error = self.assertRaises(
+            DriverLoadFailure, oslo_messaging.get_transport, self.conf)
+        self.assertIn(
+            "rpc_conn_pool_size: 1 must be greater than or equal "
+            "to conn_pool_min_size: 2", str(error))
 
 
 class TestRabbitDriverLoadSSL(test_utils.BaseTestCase):
     scenarios = [
         ('no_ssl', dict(options=dict(), expected=False)),
-        ('no_ssl_with_options', dict(options=dict(kombu_ssl_version='TLSv1'),
+        ('no_ssl_with_options', dict(options=dict(ssl_version='TLSv1'),
                                      expected=False)),
-        ('just_ssl', dict(options=dict(rabbit_use_ssl=True),
+        ('just_ssl', dict(options=dict(ssl=True),
                           expected=True)),
-        ('ssl_with_options', dict(options=dict(rabbit_use_ssl=True,
-                                               kombu_ssl_version='TLSv1',
-                                               kombu_ssl_keyfile='foo',
-                                               kombu_ssl_certfile='bar',
-                                               kombu_ssl_ca_certs='foobar'),
+        ('ssl_with_options', dict(options=dict(ssl=True,
+                                               ssl_version='TLSv1',
+                                               ssl_key_file='foo',
+                                               ssl_cert_file='bar',
+                                               ssl_ca_file='foobar'),
                                   expected=dict(ssl_version=3,
                                                 keyfile='foo',
                                                 certfile='bar',
@@ -179,7 +176,8 @@ class TestRabbitDriverLoadSSL(test_utils.BaseTestCase):
                                                 cert_reqs=ssl.CERT_REQUIRED))),
     ]
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('kombu.connection.Connection')
     def test_driver_load(self, connection_klass, fake_ensure):
         self.config(group="oslo_messaging_rabbit", **self.options)
@@ -187,13 +185,22 @@ class TestRabbitDriverLoadSSL(test_utils.BaseTestCase):
                                                  'kombu+memory:////')
         self.addCleanup(transport.cleanup)
 
-        transport._driver._get_connection()
+        connection = transport._driver._get_connection()
         connection_klass.assert_called_once_with(
-            'memory:///', transport_options={'confirm_publish': True,
-                                             'on_blocked': mock.ANY,
-                                             'on_unblocked': mock.ANY},
+            'memory:///', transport_options={
+                'client_properties': {
+                    'capabilities': {
+                        'connection.blocked': True,
+                        'consumer_cancel_notify': True,
+                        'authentication_failure_close': True,
+                    },
+                    'connection_name': connection.name},
+                'confirm_publish': True,
+                'on_blocked': mock.ANY,
+                'on_unblocked': mock.ANY},
             ssl=self.expected, login_method='AMQPLAIN',
-            heartbeat=60, failover_strategy='round-robin')
+            heartbeat=60, failover_strategy='round-robin'
+        )
 
 
 class TestRabbitPublisher(test_utils.BaseTestCase):
@@ -201,26 +208,35 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
     def test_send_with_timeout(self, fake_publish):
         transport = oslo_messaging.get_transport(self.conf,
                                                  'kombu+memory:////')
+        exchange_mock = mock.Mock()
         with transport._driver._get_connection(
                 driver_common.PURPOSE_SEND) as pool_conn:
             conn = pool_conn.connection
-            conn._publish(mock.Mock(), 'msg', routing_key='routing_key',
+            conn._publish(exchange_mock, 'msg', routing_key='routing_key',
                           timeout=1)
+
         fake_publish.assert_called_with(
             'msg', expiration=1,
-            compression=self.conf.oslo_messaging_rabbit.kombu_compression)
+            exchange=exchange_mock,
+            compression=self.conf.oslo_messaging_rabbit.kombu_compression,
+            mandatory=False,
+            routing_key='routing_key')
 
     @mock.patch('kombu.messaging.Producer.publish')
     def test_send_no_timeout(self, fake_publish):
         transport = oslo_messaging.get_transport(self.conf,
                                                  'kombu+memory:////')
+        exchange_mock = mock.Mock()
         with transport._driver._get_connection(
                 driver_common.PURPOSE_SEND) as pool_conn:
             conn = pool_conn.connection
-            conn._publish(mock.Mock(), 'msg', routing_key='routing_key')
+            conn._publish(exchange_mock, 'msg', routing_key='routing_key')
         fake_publish.assert_called_with(
             'msg', expiration=None,
-            compression=self.conf.oslo_messaging_rabbit.kombu_compression)
+            mandatory=False,
+            compression=self.conf.oslo_messaging_rabbit.kombu_compression,
+            exchange=exchange_mock,
+            routing_key='routing_key')
 
     def test_declared_queue_publisher(self):
         transport = oslo_messaging.get_transport(self.conf,
@@ -247,21 +263,42 @@ class TestRabbitPublisher(test_utils.BaseTestCase):
                     conn._publish_and_creates_default_queue,
                     exchange, {}, routing_key='foobar')
 
-            # Ensure the exchange does not exists
-            self.assertRaises(exc, try_send, e_passive)
-            # Create it
-            try_send(e_active)
-            # Ensure it creates it
-            try_send(e_passive)
+            with mock.patch('kombu.transport.virtual.Channel.close'):
+                # Ensure the exchange does not exists
+                self.assertRaises(oslo_messaging.MessageDeliveryFailure,
+                                  try_send, e_passive)
+                # Create it
+                try_send(e_active)
+                # Ensure it creates it
+                try_send(e_passive)
+
+            with mock.patch('kombu.messaging.Producer.publish',
+                            side_effect=exc):
+                with mock.patch('kombu.transport.virtual.Channel.close'):
+                    # Ensure the exchange is already in cache
+                    self.assertIn('foobar', conn._declared_exchanges)
+                    # Reset connection
+                    self.assertRaises(oslo_messaging.MessageDeliveryFailure,
+                                      try_send, e_passive)
+                    # Ensure the cache is empty
+                    self.assertEqual(0, len(conn._declared_exchanges))
 
-            with mock.patch('kombu.messaging.Producer', side_effect=exc):
-                # Should reset the cache and ensures the exchange does
-                # not exists
-                self.assertRaises(exc, try_send, e_passive)
-            # Recreate it
             try_send(e_active)
-            # Ensure it have been recreated
-            try_send(e_passive)
+            self.assertIn('foobar', conn._declared_exchanges)
+
+    def test_send_exception_remap(self):
+        bad_exc = Exception("Non-oslo.messaging exception")
+        transport = oslo_messaging.get_transport(self.conf,
+                                                 'kombu+memory:////')
+        exchange_mock = mock.Mock()
+        with transport._driver._get_connection(
+                driver_common.PURPOSE_SEND) as pool_conn:
+            conn = pool_conn.connection
+            with mock.patch('kombu.messaging.Producer.publish',
+                            side_effect=bad_exc):
+                self.assertRaises(MessageDeliveryFailure,
+                                  conn._ensure_publishing,
+                                  conn._publish, exchange_mock, 'msg')
 
 
 class TestRabbitConsume(test_utils.BaseTestCase):
@@ -287,6 +324,51 @@ class TestRabbitConsume(test_utils.BaseTestCase):
 
         self.assertEqual(0, int(deadline - time.time()))
 
+    def test_consume_from_missing_queue(self):
+        transport = oslo_messaging.get_transport(self.conf, 'kombu+memory://')
+        self.addCleanup(transport.cleanup)
+        with transport._driver._get_connection(
+                driver_common.PURPOSE_LISTEN) as conn:
+            with mock.patch('kombu.Queue.consume') as consume, mock.patch(
+                    'kombu.Queue.declare') as declare:
+                conn.declare_topic_consumer(exchange_name='test',
+                                            topic='test',
+                                            callback=lambda msg: True)
+                import amqp
+                consume.side_effect = [amqp.NotFound, None]
+                conn.connection.connection.recoverable_connection_errors = ()
+                conn.connection.connection.recoverable_channel_errors = ()
+                self.assertEqual(1, declare.call_count)
+                conn.connection.connection.drain_events = mock.Mock()
+                # Ensure that a queue will be re-declared if the consume method
+                # of kombu.Queue raise amqp.NotFound
+                conn.consume()
+                self.assertEqual(2, declare.call_count)
+
+    def test_consume_from_missing_queue_with_io_error_on_redeclaration(self):
+        transport = oslo_messaging.get_transport(self.conf, 'kombu+memory://')
+        self.addCleanup(transport.cleanup)
+        with transport._driver._get_connection(
+                driver_common.PURPOSE_LISTEN) as conn:
+            with mock.patch('kombu.Queue.consume') as consume, mock.patch(
+                    'kombu.Queue.declare') as declare:
+                conn.declare_topic_consumer(exchange_name='test',
+                                            topic='test',
+                                            callback=lambda msg: True)
+                import amqp
+                consume.side_effect = [amqp.NotFound, None]
+                declare.side_effect = [IOError, None]
+
+                conn.connection.connection.recoverable_connection_errors = (
+                    IOError,)
+                conn.connection.connection.recoverable_channel_errors = ()
+                self.assertEqual(1, declare.call_count)
+                conn.connection.connection.drain_events = mock.Mock()
+                # Ensure that a queue will be re-declared after
+                # 'queue not found' exception despite on connection error.
+                conn.consume()
+                self.assertEqual(3, declare.call_count)
+
     def test_connection_ack_have_disconnected_kombu_connection(self):
         transport = oslo_messaging.get_transport(self.conf,
                                                  'kombu+memory:////')
@@ -307,7 +389,7 @@ class TestRabbitTransportURL(test_utils.BaseTestCase):
 
     scenarios = [
         ('none', dict(url=None,
-                      expected=["amqp://guest:guest@localhost:5672//"])),
+                      expected=["amqp://guest:guest@localhost:5672/"])),
         ('memory', dict(url='kombu+memory:////',
                         expected=["memory:///"])),
         ('empty',
@@ -340,15 +422,19 @@ class TestRabbitTransportURL(test_utils.BaseTestCase):
         ('rabbit_ipv4',
          dict(url='rabbit://user:password@10.20.30.40:10/vhost',
               expected=['amqp://user:password@10.20.30.40:10/vhost'])),
+        ('rabbit_no_vhost_slash',
+         dict(url='rabbit://user:password@10.20.30.40:10',
+              expected=['amqp://user:password@10.20.30.40:10/'])),
     ]
 
     def setUp(self):
         super(TestRabbitTransportURL, self).setUp()
-        self.messaging_conf.transport_driver = 'rabbit'
+        self.messaging_conf.transport_url = 'rabbit:/'
         self.config(heartbeat_timeout_threshold=0,
                     group='oslo_messaging_rabbit')
 
-    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.ensure')
+    @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection'
+                '.ensure_connection')
     @mock.patch('oslo_messaging._drivers.impl_rabbit.Connection.reset')
     def test_transport_url(self, fake_reset, fake_ensure):
         transport = oslo_messaging.get_transport(self.conf, self.url)
@@ -380,11 +466,6 @@ class TestSendReceive(test_utils.BaseTestCase):
         ('zero', dict(rx_id=False, reply=0)),
     ]
 
-    _reply_fail = [
-        ('reply_success', dict(reply_failure_404=False)),
-        ('reply_failure', dict(reply_failure_404=True)),
-    ]
-
     _failure = [
         ('success', dict(failure=False)),
         ('failure', dict(failure=True, expected=False)),
@@ -392,8 +473,11 @@ class TestSendReceive(test_utils.BaseTestCase):
     ]
 
     _timeout = [
-        ('no_timeout', dict(timeout=None)),
-        ('timeout', dict(timeout=0.01)),  # FIXME(markmc): timeout=0 is broken?
+        ('no_timeout', dict(timeout=None, call_monitor_timeout=None)),
+        ('timeout', dict(timeout=0.01,  # FIXME(markmc): timeout=0 is broken?
+                         call_monitor_timeout=None)),
+        ('call_monitor_timeout', dict(timeout=0.01,
+                                      call_monitor_timeout=0.02)),
     ]
 
     @classmethod
@@ -401,7 +485,6 @@ class TestSendReceive(test_utils.BaseTestCase):
         cls.scenarios = testscenarios.multiply_scenarios(cls._n_senders,
                                                          cls._context,
                                                          cls._reply,
-                                                         cls._reply_fail,
                                                          cls._failure,
                                                          cls._timeout)
 
@@ -418,37 +501,31 @@ class TestSendReceive(test_utils.BaseTestCase):
 
         target = oslo_messaging.Target(topic='testtopic')
 
-        listener = driver.listen(target)
+        listener = driver.listen(target, None, None)._poll_style_listener
 
         senders = []
         replies = []
         msgs = []
-        errors = []
 
-        def stub_error(msg, *a, **kw):
-            if (a and len(a) == 1 and isinstance(a[0], dict) and a[0]):
-                a = a[0]
-            errors.append(str(msg) % a)
-
-        self.stubs.Set(driver_common.LOG, 'error', stub_error)
+        # FIXME(danms): Surely this is not the right way to do this...
+        self.ctxt['client_timeout'] = self.call_monitor_timeout
 
         def send_and_wait_for_reply(i):
             try:
-                if self.reply_failure_404:
-                    timeout = 0.01
-                else:
-                    timeout = self.timeout
+
+                timeout = self.timeout
+                cm_timeout = self.call_monitor_timeout
                 replies.append(driver.send(target,
                                            self.ctxt,
                                            {'tx_id': i},
                                            wait_for_reply=True,
-                                           timeout=timeout))
+                                           timeout=timeout,
+                                           call_monitor_timeout=cm_timeout))
                 self.assertFalse(self.failure)
                 self.assertIsNone(self.timeout)
             except (ZeroDivisionError, oslo_messaging.MessagingTimeout) as e:
                 replies.append(e)
-                self.assertTrue(self.failure or self.timeout is not None
-                                or self.reply_failure_404)
+                self.assertTrue(self.failure or self.timeout is not None)
 
         while len(senders) < self.n_senders:
             senders.append(threading.Thread(target=send_and_wait_for_reply,
@@ -468,18 +545,6 @@ class TestSendReceive(test_utils.BaseTestCase):
         if len(order) > 1:
             order[-1], order[-2] = order[-2], order[-1]
 
-        if self.reply_failure_404:
-            start = time.time()
-            # NOTE(sileht): Simulate a rpc client restart
-            # By returning a ExchangeNotFound when we try to
-            # send reply
-            exc = (driver._reply_q_conn.connection.
-                   connection.channel_errors[0]())
-            exc.code = 404
-            self.useFixture(mockpatch.Patch(
-                'kombu.messaging.Producer.publish',
-                side_effect=exc))
-
         for i in order:
             if self.timeout is None:
                 if self.failure:
@@ -487,28 +552,16 @@ class TestSendReceive(test_utils.BaseTestCase):
                         raise ZeroDivisionError
                     except Exception:
                         failure = sys.exc_info()
-                    msgs[i].reply(failure=failure,
-                                  log_failure=not self.expected)
+                    msgs[i].reply(failure=failure)
                 elif self.rx_id:
                     msgs[i].reply({'rx_id': i})
                 else:
                     msgs[i].reply(self.reply)
-            elif self.reply_failure_404:
-                msgs[i].reply({})
             senders[i].join()
 
-        if self.reply_failure_404:
-            # NOTE(sileht) all reply fail, first take
-            # kombu_missing_consumer_retry_timeout seconds to fail
-            # next immediately fail
-            dt = time.time() - start
-            rabbit_conf = self.conf.oslo_messaging_rabbit
-            timeout = rabbit_conf.kombu_missing_consumer_retry_timeout
-            self.assertTrue(timeout <= dt < (timeout + 0.100), dt)
-
         self.assertEqual(len(senders), len(replies))
         for i, reply in enumerate(replies):
-            if self.timeout is not None or self.reply_failure_404:
+            if self.timeout is not None:
                 self.assertIsInstance(reply, oslo_messaging.MessagingTimeout)
             elif self.failure:
                 self.assertIsInstance(reply, ZeroDivisionError)
@@ -517,11 +570,6 @@ class TestSendReceive(test_utils.BaseTestCase):
             else:
                 self.assertEqual(self.reply, reply)
 
-        if not self.timeout and self.failure and not self.expected:
-            self.assertTrue(len(errors) > 0, errors)
-        else:
-            self.assertEqual(0, len(errors), errors)
-
 
 TestSendReceive.generate_scenarios()
 
@@ -534,7 +582,7 @@ class TestPollAsync(test_utils.BaseTestCase):
         self.addCleanup(transport.cleanup)
         driver = transport._driver
         target = oslo_messaging.Target(topic='testtopic')
-        listener = driver.listen(target)
+        listener = driver.listen(target, None, None)._poll_style_listener
         received = listener.poll(timeout=0.050)
         self.assertEqual([], received)
 
@@ -550,8 +598,7 @@ class TestRacyWaitForReply(test_utils.BaseTestCase):
 
         target = oslo_messaging.Target(topic='testtopic')
 
-        listener = driver.listen(target)
-
+        listener = driver.listen(target, None, None)._poll_style_listener
         senders = []
         replies = []
         msgs = []
@@ -559,16 +606,18 @@ class TestRacyWaitForReply(test_utils.BaseTestCase):
         wait_conditions = []
         orig_reply_waiter = amqpdriver.ReplyWaiter.wait
 
-        def reply_waiter(self, msg_id, timeout):
+        def reply_waiter(self, msg_id, timeout, call_monitor_timeout):
             if wait_conditions:
                 cond = wait_conditions.pop()
                 with cond:
                     cond.notify()
                 with cond:
                     cond.wait()
-            return orig_reply_waiter(self, msg_id, timeout)
+            return orig_reply_waiter(self, msg_id, timeout,
+                                     call_monitor_timeout)
 
-        self.stubs.Set(amqpdriver.ReplyWaiter, 'wait', reply_waiter)
+        self.useFixture(fixtures.MockPatchObject(
+            amqpdriver.ReplyWaiter, 'wait', reply_waiter))
 
         def send_and_wait_for_reply(i, wait_for_reply):
             replies.append(driver.send(target,
@@ -622,7 +671,7 @@ class TestRacyWaitForReply(test_utils.BaseTestCase):
         # Verify the _send_reply was not invoked by driver:
         with mock.patch.object(msgs[2], '_send_reply') as method:
             msgs[2].reply({'rx_id': 2})
-            self.assertEqual(method.call_count, 0)
+            self.assertEqual(0, method.call_count)
 
         # Wait for the 3rd thread to finish
         senders[2].join()
@@ -855,11 +904,12 @@ class TestReplyWireFormat(test_utils.BaseTestCase):
     ]
 
     _context = [
-        ('empty_ctxt', dict(ctxt={}, expected_ctxt={})),
+        ('empty_ctxt', dict(ctxt={}, expected_ctxt={'client_timeout': None})),
         ('user_project_ctxt',
          dict(ctxt={'_context_user': 'mark',
                     '_context_project': 'snarkybunch'},
-              expected_ctxt={'user': 'mark', 'project': 'snarkybunch'})),
+              expected_ctxt={'user': 'mark', 'project': 'snarkybunch',
+                             'client_timeout': None})),
     ]
 
     _compression = [
@@ -887,7 +937,7 @@ class TestReplyWireFormat(test_utils.BaseTestCase):
                                        server=self.server,
                                        fanout=self.fanout)
 
-        listener = driver.listen(target)
+        listener = driver.listen(target, None, None)._poll_style_listener
 
         connection, producer = _create_producer(target)
         self.addCleanup(connection.release)
@@ -904,6 +954,7 @@ class TestReplyWireFormat(test_utils.BaseTestCase):
             '_msg_id': uuid.uuid4().hex,
             '_unique_id': uuid.uuid4().hex,
             '_reply_q': 'reply_' + uuid.uuid4().hex,
+            '_timeout': None,
         })
 
         msg['oslo.message'] = jsonutils.dumps(msg['oslo.message'])
@@ -923,27 +974,25 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
 
     def setUp(self):
         super(RpcKombuHATestCase, self).setUp()
-        self.brokers = ['host1', 'host2', 'host3', 'host4', 'host5']
-        self.config(rabbit_hosts=self.brokers,
-                    rabbit_retry_interval=0.01,
+        transport_url = 'rabbit:/host1,host2,host3,host4,host5/'
+        self.messaging_conf.transport_url = transport_url
+        self.config(rabbit_retry_interval=0.01,
                     rabbit_retry_backoff=0.01,
                     kombu_reconnect_delay=0,
                     heartbeat_timeout_threshold=0,
                     group="oslo_messaging_rabbit")
 
-        self.kombu_connect = mock.Mock()
-        self.useFixture(mockpatch.Patch(
-            'kombu.connection.Connection.connect',
-            side_effect=self.kombu_connect))
-        self.useFixture(mockpatch.Patch(
+        self.useFixture(fixtures.MockPatch(
             'kombu.connection.Connection.connection'))
-        self.useFixture(mockpatch.Patch(
+        self.useFixture(fixtures.MockPatch(
             'kombu.connection.Connection.channel'))
 
         # starting from the first broker in the list
         url = oslo_messaging.TransportURL.parse(self.conf, None)
         self.connection = rabbit_driver.Connection(self.conf, url,
                                                    driver_common.PURPOSE_SEND)
+        self.useFixture(fixtures.MockPatch(
+            'kombu.connection.Connection.connect'))
         self.addCleanup(self.connection.close)
 
     def test_ensure_four_retry(self):
@@ -951,7 +1000,6 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           self.connection.ensure, mock_callback,
                           retry=4)
-        self.assertEqual(5, self.kombu_connect.call_count)
         self.assertEqual(6, mock_callback.call_count)
 
     def test_ensure_one_retry(self):
@@ -959,7 +1007,6 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           self.connection.ensure, mock_callback,
                           retry=1)
-        self.assertEqual(2, self.kombu_connect.call_count)
         self.assertEqual(3, mock_callback.call_count)
 
     def test_ensure_no_retry(self):
@@ -967,12 +1014,12 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           self.connection.ensure, mock_callback,
                           retry=0)
-        self.assertEqual(1, self.kombu_connect.call_count)
         self.assertEqual(2, mock_callback.call_count)
 
 
 class ConnectionLockTestCase(test_utils.BaseTestCase):
     def _thread(self, lock, sleep, heartbeat=False):
+
         def thread_task():
             if heartbeat:
                 with lock.for_heartbeat():
